
from ..osascript import OSAScript, OSAObjProxy

class Item(OSAObjProxy):
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        super().__init__(script, obj_id, class_name)
        

OSAObjProxy._NAME_CLASS_MAP['item'] = Item


from .record import Record

class SmartGroup(Record):
    def __init__(self, item_ref):
        super().__init__(item_ref)
    # properties
    @property
    def exclude_subgroups(self) -> bool:
        """Exclude subgroups of the search group from searching."""
        return self.get_property_native('excludeSubgroups')

    @exclude_subgroups.setter
    def exclude_subgroups(self, value: bool):
        self.set_property('excludeSubgroups', value)

    @property
    def highlight_occurrences(self) -> bool:
        """Highlight found occurrences in documents."""
        return self.get_property_native('highlightOccurrences')

    @highlight_occurrences.setter
    def highlight_occurrences(self, value: bool):
        self.set_property('highlightOccurrences', value)

    @property
    def search_group(self) -> Record:
        """Group of the smart group to search in."""
        return self.get_property_native('searchGroup')

    @search_group.setter
    def search_group(self, value: Record):
        self.set_property('searchGroup', value)

    @property
    def search_predicates(self) -> str:
        """A string representation of the conditions of the smart group."""
        return self.get_property_native('searchPredicates')

    @search_predicates.setter
    def search_predicates(self, value: str):
        self.set_property('searchPredicates', value)



from ..osascript import OSAScript, OSAObjProxy

class Reminder(OSAObjProxy):
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        super().__init__(script, obj_id, class_name)

    # properties
    @property
    def alarm(self) -> str:
        """Alarm of reminder."""
        return self.get_property('alarm')

    @alarm.setter
    def alarm(self, value: str):
        self.set_property('alarm', value)

    @property
    def alarm_string(self) -> str:
        """Name of sound, text to speak, text of alert/notification, source/path of script or recipient of email. Text can also contain placeholders."""
        return self.get_property('alarmString')

    @alarm_string.setter
    def alarm_string(self, value: str):
        self.set_property('alarmString', value)

    @property
    def day_of_week(self) -> str:
        """Scheduled day of week."""
        return self.get_property('dayOfWeek')

    @day_of_week.setter
    def day_of_week(self, value: str):
        self.set_property('dayOfWeek', value)

    @property
    def due_date(self) -> str:
        """Due date."""
        return self.get_property('dueDate')

    @due_date.setter
    def due_date(self, value: str):
        raise NotImplementedError()

    @property
    def interval(self) -> int:
        """Interval of schedule (every n hours, days, weeks, months or years)"""
        return self.get_property('interval')

    @interval.setter
    def interval(self, value: int):
        self.set_property('interval', value)

    @property
    def masc(self) -> int:
        """Bitmap specifying scheduled days of week/month or scheduled months of year."""
        return self.get_property('masc')

    @masc.setter
    def masc(self, value: int):
        self.set_property('masc', value)

    @property
    def schedule(self) -> str:
        """Schedule of reminder."""
        return self.get_property('schedule')

    @schedule.setter
    def schedule(self, value: str):
        self.set_property('schedule', value)

    @property
    def week_of_month(self) -> str:
        """Scheduled week of month."""
        return self.get_property('weekOfMonth')

    @week_of_month.setter
    def week_of_month(self, value: str):
        self.set_property('weekOfMonth', value)


"""Not yet implemented."""
from ..osascript import OSAObjProxy

# class Text(OSAObjProxy):
#     def __init__(self, item_ref):
#         super().__init__(item_ref)

#     # elements
#     @property
#     def attachments(self) -> List[Attachment]:
#         """List of the attachments contained in the text."""
#         return self.get_element_array('attachment')

#     @property
#     def attribute_runs(self) -> List[AttributeRun]:
#         """List of attribute runs containing formatting information for portions of the text."""
#         return self.get_element_array('attributeRun')

#     @property
#     def characters(self) -> List[str]:
#         """List of the individual characters in the text."""
#         return self.get_element_array('character')

#     @property
#     def paragraphs(self) -> List[Paragraph]:
#         """List of paragraphs contained in the text."""
#         return self.get_element_array('paragraph')

#     @property
#     def words(self) -> List[Word]:
#         """List of words contained in the text."""
#         return self.get_element_array('word')

#     # properties
#     @property
#     def color(self) -> Color:
#         """The color of the first character."""
#         return self.get_property('color', Color)

#     @color.setter
#     def color(self, value: Color):
#         self.set_property('color', value)

#     @property
#     def font(self) -> Text:
#         """The name of the font of the first character."""
#         return self.get_property_native('font')

#     @font.setter
#     def font(self, value: Text):
#         self.set_property('font', value)

#     @property
#     def size(self) -> int:
#         """The size in points of the first character."""
#         return self.get_property('size', int)

#     @size.setter
#     def size(self, value: int):
#         self.set_property('size', value)

# OSAObjProxy._NAME_CLASS_MAP['text'] = Text


import os
from .osascript import OSAScript, OSAObjProxy
from .models.record import Record
from .models.database import Database
from .models.windows import (ThinkWindow, DocumentWindow, ViewerWindow)
from typing import Optional, Union, List

script_path = os.path.join(os.path.dirname(__file__), 'jxa_helper.scpt')

class DEVONthink3(OSAObjProxy):
    def __init__(self, name="DEVONthink 3") -> None:
        self.name = name
        self.script = OSAScript.from_path(script_path)
        self._ext = DevonthinkExtension(self)
        response = self.script.call_json('getApplication', {'name': self.name})
        super().__init__(self.script, response['objId'], response['className'])

    @property
    def ext(self) -> 'DevonthinkExtension':
        """Extension methods for DEVONthink 3. These methods are not part of the DEVONthink's AppleScript API.

        Examples:
            >>> dtp3 = DEVONthink3()
            >>> dtp3.ext.db_by_name('blue-book')
        Returns:
            DevonthinkExtension
        """
        return self._ext

    # elements
    @property
    def databases(self) -> List[Database]:
        return self.get_property_native('databases')
    
    @property
    def dcoument_windows(self) -> List[DocumentWindow]:
        return self.get_property_native('documentWindows')

    @property
    def selected_records(self) -> List[Record]:
        return self.get_property_native('selectedRecords')
    
    @property
    def think_windows(self) -> List[ThinkWindow]:
        return self.get_property_native('thinkWindows')

    @property
    def viewer_windows(self) -> List[ViewerWindow]:
        return self.get_property_native('viewerWindows')

    # properties
    @property
    def bates_number(self) -> int:
        """Current bates number."""
        return self.get_property_native('batesNumber')

    @property
    def cancelled_progress(self) -> bool:
        """Specifies if a process with visible progress indicator should be cancelled."""
        return self.get_property_native('cancelledProgress')

    @property
    def content_record(self) -> Record:
        """The record of the visible document in the frontmost think window."""
        return self.get_property_native('contentRecord')

    @property
    def current_database(self) -> Database:
        """The currently used database."""
        return self.get_property_native('currentDatabase')

    @property
    def current_group(self) -> Record:
        """The (selected) group of the frontmost window of the current database. Returns root of current database if no current group exists."""
        return self.get_property_native('currentGroup')

    @property
    def current_workspace(self) -> str:
        """The name of the currently used workspace."""
        return self.get_property_native('currentWorkspace')

    @property
    def inbox(self) -> Database:
        """The global inbox."""
        return self.get_property_native('inbox')

    @property
    def incoming_group(self) -> Record:
        """The default group for new notes. Either global inbox or incoming group of current database if global inbox isn't available."""
        return self.get_property_native('incomingGroup')

    @property
    def last_downloaded_URL(self) -> str:
        """The actual URL of the last download."""
        return self.get_property_native('lastDownloadedURL')

    @property
    def last_downloaded_response(self) -> Record:
        """HTTP-Status, Last-Modified, Content-Type, Content-Length and Charset of last HTTP(S) response."""
        return self.get_property_native('lastDownloadedResponse')

    @property
    def preferred_import_destination(self) -> Record:
        """The default destination for data from external sources. See Preferences > Import > Destination."""
        return self.get_property_native('preferredImportDestination')

    @property
    def reading_list(self) -> List[dict]:
        """The items of the reading list."""
        return self.get_property_native('readingList')

    @property
    def selection(self) -> list:
        """Not implemented due to the poor document. Plus it's 
        officially not recommanded:

            selection (list, r/o) : The current selection of
            the frontmost viewer window or the record of the
            frontmost document window. 'selected records'
            relationship is recommended instead especially for
            bulk retrieval of properties like UUID.

        """
        raise NotImplementedError()

    @property
    def workspaces(self) -> List[str]:
        """The names of all available workspaces."""
        return self.get_property_native('workspaces')
    
    # methods
    ## standard additions
    def display_dialog(self, text: str, place_holder: str = None, buttons: List[str]=None, default_button: str = None, with_icon: str = None) -> str:
        """Display a dialog.

        Args:
            text (str): The text to display.
            place_holder (str): The placeholder text.
            with_icon (str): The icon to display.
            buttons (List[str]): The buttons to display.
            default_button (str): The default button.

        Returns:
            str: The selected button.
        """
        kwargs = {
            'defaultAnswer': place_holder,
            'buttons': buttons,
            'defaultButton': default_button,
            'withIcon': with_icon,
        }

        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return self.run_method('displayDialog', [text], kwargs)
    
    def search(self, text: str, comparision: str = 'no case', excludeSubgroups: bool = False) -> List[Record]:
        """Search for records in specified group or all databases.

        Args:
            text (str): The search string.
            comparision (str, optional):  The comparison to use (default `'no case'`). One of fuzzy/‌no case/‌no umlauts.
            excludeSubgroups (bool, optional): Don't search in subgroups of the specified group. (default `false`).

        Returns:
            List[Record]: The list of records matching the search string.
        """
        return self.run_method('search', [text], {
            'comparision': comparision,
            'excludeSubgroups': excludeSubgroups,
        })

    def import_(self, file_path: str, source_app: str = None, record_name: str = None, placeholders: dict = None, to: Record = None) -> Record:
        """Import a file or folder (including its subfolders).

        Args:
            file_path (str): The POSIX path of the file or folder.
            source_app (str, optional): The name of the source application. Default is None.
            record_name (str, optional): The name for the imported record. Default is None.
            placeholders (dict, optional): Optional placeholders as key-value-pairs for text, RTF/RTFD & HTML/XML templates and filenames. Default is None.
            to (Group, optional): he destination group. Uses incoming group or group selector if not specified. Default is None.

        Returns:
            Record: The imported record.
        """
        return self.run_method('import', [file_path], {
            'from': source_app,
            'name': record_name,
            'placeholders': placeholders,
            'to': to,
        })

    def indicate(self, text: str, to_: Optional[Record] = None, file_type: Optional[int] = None) -> Record:
        """Indicate ('index') a file or folder (including its subfolders). If no type is specified or the type is 'all', then links to unknown file types are created too.

        Args:
            text (str): The POSIX path of the file or folder.
            dest (Record, optional): The destination group. Uses incoming group or group selector if not specified.
            file_type (int, optional): Obsolete.

        Returns:
            Record: The record of the indicated file or folder.
        """
        return self.run_method('indicate', text, {
            'to': to_,
            'fileType': file_type,
        })

    def index(self, text: str, to_: Optional[Record] = None, file_type: Optional[int] = None) -> Record:
        """Alias for `indicate`."""
        return self.indicate(text, to_, file_type)

    def open_database(self, file_path: str) -> Database:
        """Open an existing database.

        Args:
            file_path (str): The POSIX file path of the database.

        Returns:
            Database: The opened database.
        """
        return self.run_method('open database', [file_path])
    
    def create_database(self, text: str) -> Database:
        """Create a new database.

        Args:
            text (str): POSIX file path of database.

        Returns:
            Database: The new database object.
        """
        return self.run_method('createDatabase', [text])

    def create_location(self, text: str, database: Database = None) -> Record:
        """Create a hierarchy of groups if necessary.

        Args:
            text (str): The hierarchy as a POSIX path (/ in names has to be replaced with \/, see location property).
            database (Database, optional): The database. Uses current database if not specified.

        Returns:
            Record: The created record.
        """
        return self.run_method('createLocation', [text], {
            'database': database,
        })

    def create_record_with(self, properties: dict, in_: Optional['Record'] = None) -> Record:
        """Create a new record.

        Args:
            record (Record): The properties of the record. Possible keys for record are 'name', 'type', 'comment', 'path', 'URL', 'creation date', 'modification date', 'date', 'plain text', 'rich text', 'source', 'data', 'content', 'columns', 'cells', 'thumbnail' and 'tags'.
            in_ (Record, optional): The destination group for the new record. Uses incoming group or group selector if not specified.
        Returns:
            Record: The newly created record.
        """
        return self.run_method('createRecordWith', [properties], {'in': in_})

    def delete(self, record: 'Record', in_: Optional['Record'] = None) -> bool:
        """Delete all instances of a record from the database or one instance from the specified group.

        Args:
            record (Record): The record to delete.
            in_ (Record, optional): The parent group of this instance. Deletes all instances of the record from the database if not specified.

        Returns:
            bool: `True` if the deletion was successful.
        """
        return self.run_method('delete', [], {'record': record, 'in': in_})

    def get_database_with_uuid(self, text: str) -> Database:
        """Get database with the specified uuid.

        Args:
            text (str): The unique identifier.

        Returns:
            Database: The database with the specified uuid.
        """
        return self.run_method('getDatabaseWithUuid', [text])



    def move(self, record: Union[Record, List[Record]], to: Record, from_: Record = None) -> Union[Record, List[Record]]:
        """Move all instances of a record to a different group. Specify the "from" group to to move a single instance to a different group.

        Args:
            record (Union[Record, List[Record]]): The record or a list of records to move.
            to (Record): The destination group which doesn't have to be in the same database.
            from_ (Record, optional): The source group. Only applicable if record and destination group are in the same database.

        Returns:
            Union[Record, List[Record]]: The moved record or list of records.
        """
        return self.run_method('move', [], {
            'record': record,
            'to': to,
            'from': from_,
        })


    def replicate(self, record: Record, to: Record) -> Record:
        """Replicate a record.

        Args:
            record (Record): The record to replicate.
            to (Record): The destination group which must be in the same database.

        Returns:
            Record: The replicated record.
        """
        return self.run_method('replicate', [], {
            'record': record,
            'to': to,
        })

    def get_record_at(self, text: str, database: Optional[Database] = None) -> Record:
        """Search for record at the specified location.

        Args:
            text (str): The location of the record as a POSIX path (/ in names has to be replaced with \/, see location property).
            database (Database, optional): The database. Uses current database if not specified.

        Returns:
            Record: The record at the specified location.
        """
        return self.run_method('getRecordAt', [text], {'in': database})

    def get_record_with_uuid(self, text: str, database: Optional[Database] = None) -> Record:
        """Get record with the specified uuid or item link.

        Args:
            text (str): The unique identifier or item link.
            database (Database, optional): The database. Uses all databases if not specified.

        Returns:
            Record: The record with the specified uuid or item link.
        """
        return self.run_method('getRecordWithUuid', [text], {'in': database})

class DevonthinkExtension:
    def __init__(self, app: DEVONthink3):
        self.app = app
    
    def db_by_name(self, name: str) -> Optional[Database]:
        dbs = self.app.databases
        for db in dbs:
            if db.name == name:
                return db



from typing import List
from ..osascript import OSAScript, OSAObjProxy
from .record import Record
from .database import Database

class ThinkWindow(OSAObjProxy):
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        super().__init__(script, obj_id, class_name)

    # elements
    @property
    def contains_tabs(self) -> bool:
        """Whether the window contains tabs"""
        return True

    # properties
    @property
    def content_record(self) -> Record:
        """The record of the visible document."""
        return self.get_property_native('contentRecord')

    @property
    def current_line(self) -> int:
        """Zero-based index of current line."""
        return self.get_property_native('currentLine')

    @property
    def current_movie_frame(self):
        """Thumbnail of current movie frame."""
        raise NotImplementedError()

    @property
    def current_page(self) -> int:
        """Zero-based index of current PDF page."""
        return self.get_property_native('currentPage')

    @current_page.setter
    def current_page(self, value: int):
        self.set_property('currentPage', value)

    @property
    def current_tab(self) -> None:
        """The selected tab of the think window."""
        raise NotImplementedError()

    @property
    def current_time(self) -> float:
        """Time of current movie frame."""
        return self.get_property_native('currentTime')

    @property
    def database(self) -> Database:
        """The database of the window."""
        return self.get_property_native('database')

    @property
    def loading(self) -> bool:
        """Specifies if the current web page is still loading."""
        return self.get_property_native('loading')

    @property
    def number_of_columns(self) -> int:
        """Number of columns of the current sheet."""
        return self.get_property_native('numberOfColumns')

    @property
    def number_of_rows(self) -> int:
        """Number of rows of the current sheet."""
        return self.get_property_native('numberOfRows')

    @property
    def paginated_pdf(self) -> str:
        """A printed PDF with pagination of the visible document."""
        return self.get_property_native('paginatedPDF')

    @property
    def pdf(self) -> str:
        """A PDF without pagination of the visible document retaining the screen layout."""
        return self.get_property_native('pdf')

    @property
    def selected_column(self) -> int:
        """Index (1...n) of selected column of the current sheet."""
        return self.get_property_native('selectedColumn')

    @selected_column.setter
    def selected_column(self, value: int):
        self.set_property('selectedColumn', value)

    @property
    def selected_columns(self) -> List[int]:
        """Indices (1...n) of selected columns of the current sheet."""
        return self.get_property_native('selectedColumns')

    @property
    def selected_row(self) -> int:
        """Index (1...n) of selected row of the current sheet."""
        return self.get_property_native('selectedRow')

    @selected_row.setter
    def selected_row(self, value: int):
        self.set_property('selectedRow', value)

    @property
    def selected_rows(self) -> List[int]:
        """Indices (1...n) of selected rows of the current sheet."""
        return self.get_property_native('selectedRows')

    @property
    def selected_text(self) -> str:
        """The text container for the selection of the window."""
        return self.get_property_native('selectedText')
    
    @selected_text.setter
    def selected_text(self, value: str):
        self.set_property('selectedText', value)
    
    @property
    def source(self) -> str:
        """The HTML source of the current web page."""
        return self.get_property_native('source')

    @property
    def text(self) -> str:
        """The text container of the window."""
        return self.get_property_native('text')

    @property
    def url(self) -> str:
        """The URL of the current web page. In addition, setting the URL can be used to load a web page."""
        return self.get_property_native('url')

    @url.setter
    def url(self, value: str):
        self.set_property('url', value)

    @property
    def web_archive(self) -> str:
        """Web archive of the current web page."""
        return self.get_property_native('webArchive')

OSAObjProxy._NAME_CLASS_MAP['thinkWindow'] = ThinkWindow

class DocumentWindow(ThinkWindow):
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        super().__init__(script, obj_id, class_name)
    
    @property
    def record(self) -> Record:
        """The record of the visible document."""
        return self.get_property_native('record')

OSAObjProxy._NAME_CLASS_MAP['documentWindow'] = ThinkWindow


class ViewerWindow(ThinkWindow):
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        super().__init__(script, obj_id, class_name)
    # elements
    @property
    def selected_records(self) -> List[Record]:
        """List of selected records in the viewer."""
        return self.get_element_array('selectedRecords')

    # properties
    @property
    def root(self) -> Record:
        """The top level group of the viewer."""
        return self.get_property_native('root')

    @root.setter
    def root(self, value: Record):
        self.set_property('root', value)

    @property
    def search_query(self) -> str:
        """The search query. Setting the query performs a search."""
        return self.get_property_native('searchQuery')

    @search_query.setter
    def search_query(self, value: str):
        self.set_property('searchQuery', value)

    @property
    def search_results(self) -> list:
        """The search results."""
        return self.get_property_native('searchResults')

    @property
    def selection(self) -> list:
        """The current selection."""
        raise NotImplementedError()
OSAObjProxy._NAME_CLASS_MAP['viewerWindow'] = ThinkWindow




from ..osascript import OSAScript, OSAObjProxy
from typing import List
from .record import Record
from .smartgroup import SmartGroup

class Database(OSAObjProxy):
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        super().__init__(script, obj_id, class_name)
        
    # elements
    @property
    def contents(self) -> List['Record']:
        """The contents of the database."""
        return self.get_property_native('contents')

    @property
    def parents(self) -> List['Record']:
        """The parents of the database."""
        return self.get_property_native('parents')

    @property
    def records(self) -> List['Record']:
        """The records contained in the database."""
        return self.get_property_native('records')

    @property
    def smart_groups(self) -> List['SmartGroup']:
        """The smart groups contained in the database."""
        return self.get_property_native('smartGroups')

    # @property
    # def tag_groups(self) -> List[TagGroup]:
    #     """The tag groups contained in the database."""
    #     return self.get_property_native('tagGroups')

    # properties
    @property
    def annotations_group(self) -> 'Record':
        """The group for annotations, will be created if necessary."""
        return self.get_property_native('annotationsGroup')

    @property
    def comment(self) -> str:
        """The comment of the database."""
        return self.get_property_native('comment')
    
    @comment.setter
    def comment(self, value: str):
        self.set_property('comment', value)

    @property
    def current_group(self) -> 'Record':
        """The (selected) group of the frontmost window. Returns root if no current group exists."""
        return self.get_property_native('currentGroup')

    @property
    def encrypted(self) -> bool:
        """Specifies if a database is encrypted or not."""
        return self.get_property_native('encrypted')

    @property
    def id(self) -> int:
        """The scripting identifier of a database."""
        return self.get_property_native('id')

    @property
    def incoming_group(self) -> 'Record':
        """The default group for new notes. Might be identical to root."""
        return self.get_property_native('incomingGroup')

    @property
    def name(self) -> str:
        """The name of the database."""
        return self.get_property_native('name')
    
    @name.setter
    def name(self, value: str):
        self.set_property('name', value)

    @property
    def path(self) -> str:
        """The POSIX path of the database."""
        return self.get_property_native('path')

    @property
    def read_only(self) -> bool:
        """Specifies if a database is read-only and can't be modified."""
        return self.get_property_native('readOnly')

    @property
    def root(self) -> 'Record':
        """The top level group of the database."""
        return self.get_property_native('root')

    @property
    def sync_group(self) -> 'Record':
        """Obsolete group for synchronizing with DEVONthink To Go 1.x."""
        raise NotImplementedError()

    @property
    def tags_group(self) -> 'Record':
        """The group for tags."""
        return self.get_property_native('tagsGroup')

    @property
    def trash_group(self) -> 'Record':
        """The trash's group."""
        return self.get_property_native('trashGroup')

    @property
    def uuid(self) -> str:
        """The unique and persistent identifier of a database for external referencing."""
        return self.get_property_native('uuid')


OSAObjProxy._NAME_CLASS_MAP['database'] = Database


import json
from logging import getLogger
from typing import Any
import weakref
from Foundation import NSAppleScript, NSURL, NSAppleEventDescriptor
logger = getLogger(__name__)

class _PyObjectSpecifier:
    def __init__(self, script: 'OSAScript', obj_id: int):
        self.script = script
        self.obj_id = obj_id
    
    def __call__(self, args=None, kwargs=None) -> Any:
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        response  = self.script.call_json('callSelf', {
            'objId': self.obj_id,
            'args': [self.script.pyobj_to_json(arg) for arg in args],
            'kwargs': {k: self.script.pyobj_to_json(v) for k, v in kwargs.items()}
        })

        return self.json_to_pyobj(self.script, response)
    
    @classmethod
    def json_to_pyobj(cls, script: 'OSAScript', response: dict):
        if response['type'] == 'value':
            return response['data']
        elif response['type'] == 'reference':
            obj_id = response['objId']
            return cls(script, obj_id)
        elif response['type'] == 'container':
            data = response['data']
            if isinstance(data, list):
                return [
                    cls.json_to_pyobj(script, i) for i in data
                ]
            elif isinstance(data, dict):
                return {
                    k: cls.json_to_pyobj(script, v) for k, v in data.items()
                }

    def __getattr__(self, name: str):
        pass 
    
    
class OSAObjProxy:
    _NAME_CLASS_MAP = {}
    _cached_proxies = {} # {script: {obj_id: proxy}}
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        self.script = script
        self.obj_id = obj_id
        self.class_name = class_name

    def get_property_raw(self, name: str):
        response = self.script.get_properties(self.obj_id, [name])[name]
        return response

    def set_property(self, name: str, value):
        self.script.set_properties_values(self.obj_id, {name: value})

    def get_property_native(self, name: str):
        return self.json_to_pyobj(self.script, self.get_property_raw(name))
    
    def run_method(self, name: str, args, kwargs: dict = None):
        if kwargs is None:
            kwargs = {}

        kwargs = {k: self.pyobj_to_json(v) for k, v in kwargs.items()}
        args = [self.pyobj_to_json(arg) for arg in args]

        params = {'objId': self.obj_id, 'name': name, 'args': args, 'kwargs': kwargs}
        logger.debug(f'run_method params: {params}')
        response = self.script.call_json('runMethod', params)

        logger.debug(f'run_method response: {response}')

        return self.json_to_pyobj(self.script, response)

    @classmethod
    def json_to_pyobj(cls, script: 'OSAScript', response: dict):
        if response['type'] == 'value':
            return response['data']
        elif response['type'] == 'reference':
            class_name = response['className']
            obj_id = response['objId']
            # Cache the proxy object to avoid creating multiple proxies for the same object
            cached_proxy_ref = cls._cached_proxies.get(id(script), {}).get(obj_id, None)
            if cached_proxy_ref is None:
                reference_cls = cls._NAME_CLASS_MAP.get(class_name, DefaultOSAObjProxy)
                cached_proxy = reference_cls(script, obj_id, class_name)
                cls._cached_proxies.setdefault(id(script), {})[obj_id] = weakref.ref(cached_proxy)
            else:
                cached_proxy = cached_proxy_ref()
            return cached_proxy
        elif response['type'] == 'container':
            data = response['data']
            if isinstance(data, list):
                return [
                    cls.json_to_pyobj(script, i) for i in data
                ]
            elif isinstance(data, dict):
                return {
                    k: cls.json_to_pyobj(script, v) for k, v in data.items()
                }
    
    @classmethod
    def pyobj_to_json(cls, obj):
        if isinstance(obj, (int, float, str, bool, type(None))):
            return {
                'type': 'value',
                'data': obj
            }
        elif isinstance(obj, (list, tuple)):
            return {
                'type': 'container',
                'data': [cls.pyobj_to_json(i) for i in obj]
            }
        elif isinstance(obj, dict):
            return {
                'type': 'container',
                'data': {k: cls.pyobj_to_json(v) for k, v in obj.items()}
            }
        elif isinstance(obj, OSAObjProxy):
            return {
                'type': 'reference',
                'objId': obj.obj_id
            }
        else:
            raise TypeError(f'Unsupported type: {type(obj)}')
    
    def __del__(self):
        logger.debug(f'Releasing object: {self.obj_id} {self.class_name} {self} {id(self)}')
        self.script.call_json('releaseObject', {'objId': self.obj_id})


class DefaultOSAObjProxy(OSAObjProxy):
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        super().__init__(script, obj_id, class_name)
    
    def __getattr__(self, name):
        return self.get_property_native(name)

class OSAScript:
    def __init__(self, script: NSAppleScript):
        self.script = script

    @classmethod
    def from_path(cls, path):
        url = NSURL.fileURLWithPath_(path)
        script, error = NSAppleScript.alloc().initWithContentsOfURL_error_(url, None)
        if error:
            raise RuntimeError(error)
        return cls(script)
    
    def call(self, func_name: str, param: str):
        script = self.script
        event = NSAppleEventDescriptor.appleEventWithEventClass_eventID_targetDescriptor_returnID_transactionID_(
            self.fourcharcode(b'ascr'), self.fourcharcode(b'psbr'), NSAppleEventDescriptor.nullDescriptor(), 0, 0)
        descriptor_list = NSAppleEventDescriptor.listDescriptor()
        descriptor_list.insertDescriptor_atIndex_(NSAppleEventDescriptor.descriptorWithString_(param), 0)
        event.setDescriptor_forKeyword_(descriptor_list, self.fourcharcode(b'----'))

        event.setDescriptor_forKeyword_(NSAppleEventDescriptor.descriptorWithString_(func_name), self.fourcharcode(b'snam'))

        result, error = script.executeAppleEvent_error_(event, None)
        if error:
            raise RuntimeError(error)
        else:
            return result.stringValue()
    
    def call_json(self, func_name: str, param: dict):
        param_str = json.dumps(param)
        logger.debug(f'call_json func_name {func_name} param: {param_str}')
        result = self.call(func_name, param_str)
        logger.debug(f'call_json result: {result}')
        return json.loads(result)
    
    def get_properties(self, obj_id: int, names: list):
        params = {'objId': obj_id, 'names': names}
        response = self.call_json('getProperties', params)
        return response
    
    def set_properties_values(self, obj_id: int, properties: dict):
        params = {'objId': obj_id, 'properties': properties}
        self.call_json('setPropertyValues', params)
        logger.debug(f'set_properties_values params: {params}')

    
    def fourcharcode(self, chars: bytes):
        return int.from_bytes(chars, 'big')

if __name__ == '__main__':
    script = OSAScript.from_path('/Users/koc/Developer/devonthink/python-api/pydt3/test.scpt')
    print(script.call('echo', 'hello world'))



from typing import List
from ..osascript import OSAScript, OSAObjProxy

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .database import Database
    from .reminder import Reminder

class Record(OSAObjProxy):
    def __init__(self, script: 'OSAScript', obj_id: int, class_name: str):
        super().__init__(script, obj_id, class_name)

    # elements
    @property
    def children(self) -> List['Record']:
        """List of child records."""
        return self.get_element_children('children')

    @property
    def incoming_references(self) -> List['Record']:
        """List of records referencing this record."""
        return self.get_element_children('incomingReferences')

    @property
    def incoming_wiki_references(self) -> List['Record']:
        """List of wiki records referencing this record."""
        return self.get_element_children('incomingWikiReferences')

    @property
    def outgoing_references(self) -> List['Record']:
        """List of records this record references."""
        return self.get_element_children('outgoingReferences')

    @property
    def outgoing_wiki_references(self) -> List['Record']:
        """List of wiki records this record references."""
        return self.get_element_children('outgoingWikiReferences')

    @property
    def parents(self) -> List['Record']:
        """List of parent records."""
        return self.get_element_children('parents')

    @property
    def texts(self):
        """List of texts of the record."""
        raise NotImplementedError()  

    # properties
    @property
    def addition_date(self) -> str:
        """Date when the record was added to the database."""
        return self.get_property_native('additionDate')

    @property
    def aliases(self) -> str:
        """Wiki aliases (separated by commas or semicolons) of a record."""
        return self.get_property_native('aliases')

    @aliases.setter
    def aliases(self, value: str):
        self.set_property('aliases', value)

    @property
    def all_document_dates(self) -> List[str]:
        """All dates extracted from text of document, e.g. a scan."""
        return self.get_property_native('allDocumentDates')

    @property
    def altitude(self) -> float:
        """The altitude in metres of a record."""
        return self.get_property_native('altitude')

    @altitude.setter
    def altitude(self, value: float):
        self.set_property('altitude', value)

    @property
    def annotation(self) -> 'Record':
        """Annotation of a record. Only plain & rich text and Markdown documents are supported."""
        return self.get_property_native('annotation')

    @annotation.setter
    def annotation(self, value: 'Record'):
        self.set_property('annotation', value)

    @property
    def annotation_count(self) -> int:
        """The number of annotations. Currently only supported for PDF documents."""
        return self.get_property_native('annotationCount')

    @property
    def attached_script(self) -> str:
        """POSIX path of script attached to a record."""
        return self.get_property_native('attachedScript')

    @attached_script.setter
    def attached_script(self, value: str):
        self.set_property('attachedScript', value)

    @property
    def attachment_count(self) -> int:
        """The number of attachments. Currently only supported for RTFD documents and emails."""
        return self.get_property_native('attachmentCount')

    @property
    def bates_number(self) -> int:
        """Bates number."""
        return self.get_property_native('batesNumber')

    @bates_number.setter
    def bates_number(self, value: int):
        self.set_property('batesNumber', value)
        
    @property
    def cells(self) -> list:
        """The cells of a sheet. This is a list of rows, each row contains a list of string values for the various colums."""
        return self.get_property_native('cells')

    @property
    def character_count(self) -> int:
        """The character count of a record."""
        return self.get_property_native('characterCount')

    @property
    def color(self) -> List[float]:
        """The color of a record. Currently only supported by tags."""
        return self.get_property_native('color')

    @color.setter
    def color(self, value: List[float]):
        "There seems to be a bug when setting the color "
        raise NotImplementedError()

    @property
    def columns(self) -> List[str]:
        """The column names of a sheet.""" 
        return self.get_property_native('columns')

    @property
    def comment(self) -> str:
        """The comment of a record."""
        return self.get_property_native('comment')

    @comment.setter
    def comment(self, value: str):
        self.set_property('comment', value)

    @property
    def content_hash(self) -> str:
        """Stored SHA1 hash of files and document packages."""
        return self.get_property_native('contentHash')

    @property
    def creation_date(self) -> str:
        """The creation date of a record."""
        return self.get_property_native('creationDate')

    @property
    def custom_meta_data(self) -> 'Record':
        """User-defined metadata of a record as a dictionary containing key-value pairs. Setting a value for an unknown key automatically adds a definition to Preferences > Data."""
        return self.get_property_native('customMetaData')

    @property
    def data(self) -> str:
        """The file data of a record. Currently only supported by PDF documents, images, rich texts and web archives."""
        return self.get_property_native('data')

    @property
    def database(self) -> 'Database':
        """The database of the record."""
        return self.get_property_native('database')

    @property
    def date(self) -> str:
        """The (creation/modification) date of a record."""
        return self.get_property_native('date')

    @property
    def digital_object_identifier(self) -> str:
        """Digital object identifier (DOI) extracted from text of document, e.g. a scanned receipt."""
        return self.get_property_native('digitalObjectIdentifier')

    @property
    def dimensions(self) -> List[int]:
        """The width and height of an image or PDF document in pixels."""
        return self.get_property_native('dimensions')

    @property
    def document_amount(self) -> str:
        """Amount extracted from text of document, e.g. a scanned receipt."""
        return self.get_property_native('documentAmount')

    @property
    def document_date(self) -> str:
        """First date extracted from text of document, e.g. a scan."""
        return self.get_property_native('documentDate')

    @property
    def document_name(self) -> str:
        """Name based on text or properties of document"""
        return self.get_property_native('documentName')

    @property
    def dpi(self) -> int:
        """The resultion of an image in dpi."""
        return self.get_property_native('dpi')

    @property
    def duplicates(self) -> List['Record']:
        """The duplicates of a record (only other instances, not including the record)."""
        return self.get_property_object_list('duplicates')

    @property
    def duration(self) -> float:
        """The duration of audio and video files."""
        return self.get_property_native('duration')

    @property
    def encrypted(self) -> bool:
        """Specifies if a document is encrypted or not. Currently only supported by PDF documents."""
        return self.get_property_native('encrypted')

    @property
    def filename(self) -> str:
        """The current filename of a record."""
        return self.get_property_native('filename')

    @property
    def exclude_from_classification(self) -> bool:
        """Exclude group or record from classifying."""
        return self.get_property_native('excludeFromClassification')

    @exclude_from_classification.setter
    def exclude_from_classification(self, value: bool):
        self.set_property('excludeFromClassification', value)

    @property
    def exclude_from_search(self) -> bool:
        """Exclude group or record from searching."""
        return self.get_property_native('excludeFromSearch')

    @exclude_from_search.setter
    def exclude_from_search(self, value: bool):
        self.set_property('excludeFromSearch', value)

    @property
    def exclude_from_see_also(self) -> bool:
        """Exclude record from see also."""
        return self.get_property_native('excludeFromSeeAlso')

    @exclude_from_see_also.setter
    def exclude_from_see_also(self, value: bool):
        self.set_property('excludeFromSeeAlso', value)

    @property
    def exclude_from_tagging(self) -> bool:
        """Exclude group from tagging."""
        return self.get_property_native('excludeFromTagging')

    @exclude_from_tagging.setter
    def exclude_from_tagging(self, value: bool):
        self.set_property('excludeFromTagging', value)

    @property
    def exclude_from_wiki_linking(self) -> bool:
        """Exclude record from automatic Wiki linking."""
        return self.get_property_native('excludeFromWikiLinking')

    @exclude_from_wiki_linking.setter
    def exclude_from_wiki_linking(self, value: bool):
        self.set_property('excludeFromWikiLinking', value)

    @property
    def filename(self) -> str:
        """The current filename of a record."""
        return self.get_property_native('filename')

    @property
    def geolocation(self) -> str:
        """The human readable geogr. place of a record."""
        return self.get_property_native('geolocation')

    @property
    def height(self) -> int:
        """The height of an image or PDF document in pixels."""
        return self.get_property_native('height')

    @property
    def id(self) -> int:
        """The scripting identifier of a record. Optimizing or closing a database might modify this identifier."""
        return self.get_property_native('id')

    # properties
    @property
    def image(self):
        """The image or PDF document of a record. Setting supports both raw data and strings containing paths or URLs."""
        return self.get_property_native('image')

    @image.setter
    def image(self, value):
        self.set_property('image', value)

    @property
    def indexed(self) -> bool:
        """Indexed or imported record."""
        return self.get_property_native('indexed')

    @property
    def interval(self) -> float:
        """Refresh interval of a feed. Currently overriden by preferences."""
        return self.get_property_native('interval')

    @interval.setter
    def interval(self, value: float):
        self.set_property('interval', value)

    @property
    def kind(self) -> str:
        """The human readable and localized kind of a record. WARNING: Don't use this to check the type of a record, otherwise your script might fail depending on the version and the localization."""
        return self.get_property_native('kind')

    @property
    def label(self) -> int:
        """Index of label (0-7) of a record."""
        return self.get_property_native('label')

    @label.setter
    def label(self, value: int):
        self.set_property('label', value)
    @property
    def latitude(self) -> float:
        """The latitude in degrees of a record."""
        return self.get_property_native('latitude')

    @latitude.setter
    def latitude(self, value: float):
        self.set_property('latitude', value)

    @property
    def location(self) -> str:
        """The primary location in the database as a POSIX path (/ in names is replaced with \/). 
        This is the location of the first parent group."""
        return self.get_property_native('location')

    @property
    def location_group(self) -> 'Record':
        """The group of the record's primary location. This is identical to the first parent group."""
        return self.get_property_native('locationGroup')

    @property
    def locking(self) -> bool:
        """The locking of a record."""
        return self.get_property_native('locking')

    @locking.setter
    def locking(self, value: bool):
        self.set_property('locking', value)

    @property
    def longitude(self) -> float:
        """The longitude in degrees of a record."""
        return self.get_property_native('longitude')

    @longitude.setter
    def longitude(self, value: float):
        self.set_property('longitude', value)

    @property
    def meta_data(self) -> dict:
        """Document metadata (e.g. of PDF or RTF) of a record as a dictionary containing key-value pairs. 
        Possible keys are currently kMDItemTitle, kMDItemHeadline, kMDItemSubject, kMDItemDescription, 
        kMDItemCopyright, kMDItemComment, kMDItemURL, kMDItemKeywords, kM"""
        return self.get_property_native('metaData')

    @property
    def mime_type(self) -> str:
        """The (proposed) MIME type of a record."""
        return self.get_property_native('mimeType')

    @property
    def modification_date(self) -> str:
        """The modification date of a record."""
        return self.get_property_native('modificationDate')

    @modification_date.setter
    def modification_date(self, value: str):
        raise NotImplementedError()

    @property
    def name(self) -> str:
        """The name of a record."""
        return self.get_property_native('name')

    @name.setter
    def name(self, value: str):
        self.set_property('name', value)

    @property
    def name_without_date(self) -> str:
        """The name of a record without any dates."""
        return self.get_property_native('nameWithoutDate')

    @property
    def name_without_extension(self) -> str:
        """The name of a record without a file extension (independent of preferences)."""
        return self.get_property_native('nameWithoutExtension')

    @property
    def newest_document_date(self) -> date:
        """Newest date extracted from text of document, e.g. a scan."""
        return self.get_property_native('newestDocumentDate')

    @property
    def number_of_duplicates(self) -> int:
        """The number of duplicates of a record."""
        return self.get_property_native('numberOfDuplicates')

    @property
    def number_of_hits(self) -> int:
        """The number of hits of a record."""
        return self.get_property_native('numberOfHits')

    @number_of_hits.setter
    def number_of_hits(self, value: int):
        self.set_property('numberOfHits', value)

    @property
    def number_of_replicants(self) -> int:
        """The number of replicants of a record."""
        return self.get_property_native('numberOfReplicants')

    @property
    def oldest_document_date(self) -> date:
        """Oldest date extracted from text of document, e.g. a scan."""
        return self.get_property_native('oldestDocumentDate')

    @property
    def opening_date(self) -> date:
        """Date when a content was opened the last time or when a feed was refreshed the last time."""
        return self.get_property_native('openingDate')

    @property
    def page_count(self) -> int:
        """The page count of a record. Currently only supported by PDF documents."""
        return self.get_property_native('pageCount')

    @property
    def paginated_pdf(self) -> any:
        """A printed/converted PDF of the record."""
        return self.get_property_native('paginatedPDF')

    @property
    def path(self) -> str:
        """The POSIX file path of a record. Only the path of external records can be changed."""
        return self.get_property_native('path')
    
    @path.setter
    def path(self, value: str):
        self.set_property('path', value)

    @property
    def pending(self) -> bool:
        """Flag whether the (latest) contents of a record haven't been downloaded from a sync location yet."""
        return self.get_property_native('pending')

    @property
    def plain_text(self) -> str:
        """The plain text of a record."""
        return self.get_property_native('plainText')

    @property
    def proposed_filename(self) -> str:
        """The proposed filename for a record."""
        return self.get_property_native('proposedFilename')

    @property
    def rating(self) -> int:
        """Rating (0-5) of a record."""
        return self.get_property_native('rating')

    @rating.setter
    def rating(self, value: int):
        self.set_property('rating', value)

    @property
    def reference_url(self) -> str:
        """The URL (x-devonthink-item://...) to reference/link back to a record.
        Append ?page= to specify the zero-based index of a page of a PDF document,
        ?time= to specify the time of a movie or ?search= to specify a string to search.
        """
        return self.get_property_native('referenceURL')

    @property
    def reminder(self) -> 'Reminder':
        """Reminder of a record."""
        return self.get_property_native('reminder')

    @reminder.setter
    def reminder(self, value: 'Reminder'):
        self.set_property('reminder', value)

    @property
    def rich_text(self) -> str:
        """The rich text of a record (see text suite). 
        Use the 'text' relationship introduced by version 3.0 instead,
        especially for changing the contents/styles of RTF(D) documents."""
        return self.get_property_native('richText')

    @property
    def score(self) -> float:
        """The score of the last comparison, classification or search (value between 0.0 and 1.0) or undefined otherwise."""
        return self.get_property_native('score')

    @property
    def size(self) -> int:
        """The size of a record in bytes."""
        return self.get_property_native('size')

    @property
    def source(self) -> str:
        """The HTML/XML source of a record if available or the record converted to HTML if possible."""
        return self.get_property_native('source')

    @property
    def state(self) -> bool:
        """The state/flag of a record."""
        return self.get_property_native('state')

    @state.setter
    def state(self, value: bool):
        self.set_property('state', value)

    @property
    def state_visibility(self) -> bool:
        """Obsolete."""
        raise NotImplementedError()

    @property
    def tag_type(self) -> str:
        """The tag type of a record."""
        return self.get_property_native('tagType')

    @property
    def tags(self) -> List[str]:
        """The tags of a record."""
        return self.get_property_native('tags')

    @tags.setter
    def tags(self, value: list):
        self.set_property('tags', value)

    @property
    def thumbnail(self) -> str:
        """The thumbnail of a record. Setting supports both raw data and strings containing paths or URLs."""
        return self.get_property_native('thumbnail')

    @thumbnail.setter
    def thumbnail(self, value: any):
        self.set_property('thumbnail', value)

    @property
    def type(self) -> str:
        """
        The type of a record. ("bookmark"/‌"feed"/‌"formatted note"/‌"group"/‌"html"/‌"markdown"/‌"PDF document"/‌"picture"/‌"plist"/‌"quicktime"/‌"rtf"/‌"rtfd"/‌"script"/‌"sheet"/‌"smart group"/‌"txt"/‌"unknown"/‌"webarchive"/‌"xml")
        Note: In compiled menu/toolbar scripts you might have to use a string representation of the type for comparisons.
        """
        return self.get_property_native('type')

    @property
    def unread(self) -> bool:
        """The unread flag of a record."""
        return self.get_property_native('unread')

    @unread.setter
    def unread(self, value: bool):
        self.set_property('unread', value)

    @property
    def url(self) -> str:
        """The URL of a record."""
        return self.get_property_native('url')

    @property
    def uuid(self) -> str:
        """The unique and persistent identifier of a record."""
        return self.get_property_native('uuid')

    @property
    def web_archive(self) -> any:
        """The web archive of a record if available or the record converted to web archive if possible."""
        return self.get_property_native('webArchive')

    @property
    def width(self) -> int:
        """The width of an image or PDF document in pixels."""
        return self.get_property_native('width')

    @property
    def word_count(self) -> int:
        """The word count of a record."""
        return self.get_property_native('wordCount')

OSAObjProxy._NAME_CLASS_MAP['record'] = Record
OSAObjProxy._NAME_CLASS_MAP['content'] = Record
OSAObjProxy._NAME_CLASS_MAP['parent'] = Record
OSAObjProxy._NAME_CLASS_MAP['selectedRecord'] = Record